###
# Copyright 2017 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###


# -*- coding: utf-8 -*-
"""This is the helper module for RDMC"""

# ---------Imports---------
from __future__ import unicode_literals

import json
import logging
import os
import sys
import time
from collections import OrderedDict
from ctypes import byref, c_char_p, create_string_buffer

import pyaes
import six
from prompt_toolkit.completion import Completer, Completion

import redfish.hpilo.risblobstore2 as risblobstore2
import redfish.ris

try:
    import versioning
except ImportError:
    from ilorest import versioning

if os.name == "nt":
    from six.moves import winreg

    if six.PY2:
        from win32con import HKEY_LOCAL_MACHINE
    elif six.PY3:
        from win32.lib.win32con import HKEY_LOCAL_MACHINE

# ---------End of imports---------


# ---------Debug logger---------
# Using hard coded list until better solution is found
HARDCODEDLIST = [
    "modified",
    "attributeregistry",
    "links",
    "settingsresult",
    "actions",
    "availableactions",
    "extref",
]


class InfoFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno in (logging.DEBUG, logging.INFO, logging.WARN)


LOGGER = logging.getLogger()
# default logging level setting
LOGGER.setLevel(logging.ERROR)
# loggin format
LFMT = logging.Formatter("%(levelname)s\t: %(message)s")
# when success, all logs should be in stdout. When error, it should goto stderr
# Initialize StreamHandler for Stdout and Stderr logs
LOUT = logging.StreamHandler(sys.stdout)
LERR = logging.StreamHandler(sys.stderr)
# set formatter
LERR.setFormatter(LFMT)
# default stderr level setting
LERR.setLevel(logging.ERROR)
# add logger handle
LOGGER.addHandler(LERR)
LOUT.setFormatter(LFMT)
# default stdout level setting
LOUT.setLevel(logging.WARN)
LOUT.addFilter(InfoFilter())
# add logger handle
LOGGER.addHandler(LOUT)


# ---------End of debug logger---------


class ReturnCodes(object):
    """Return code class to be used by all functions"""

    SUCCESS = 0

    # ****** RDMC ERRORS ******
    CONFIGURATION_FILE_ERROR = 1
    COMMAND_NOT_ENABLED_ERROR = 2
    INVALID_COMMAND_LINE_ERROR = 3
    INVALID_FILE_FORMATTING_ERROR = 4
    USER_NOT_ADMIN = 5
    NO_CONTENTS_FOUND_FOR_OPERATION = 6
    INVALID_FILE_INPUT_ERROR = 7
    NO_CHANGES_MADE_OR_FOUND = 8
    NO_VALID_INFO_ERROR = 9

    # ****** CLI ERRORS ******
    UI_CLI_ERROR_EXCEPTION = 10
    UI_CLI_WARN_EXCEPTION = 11
    UI_CLI_USAGE_EXCEPTION = 12
    UI_CLI_COMMAND_NOT_FOUND_EXCEPTION = 13
    INVALID_PASSWORD_LENGTH_ERROR = 14

    # ****** RMC/RIS ERRORS ******
    RIS_UNDEFINED_CLIENT_ERROR = 21
    RIS_CURRENTLY_LOGGED_IN_ERROR = 22
    RIS_INSTANCE_NOT_FOUND_ERROR = 23
    RIS_NOTHING_SELECTED_ERROR = 24
    RIS_NOTHING_SELECTED_FILTER_ERROR = 25
    RIS_NOTHING_SELECTED_SET_ERROR = 26
    RIS_INVALID_SELECTION_ERROR = 27
    RIS_VALIDATION_ERROR = 28
    RIS_MISSING_ID_TOKEN = 29
    RIS_SESSION_EXPIRED = 30

    # ****** REST V1 ERRORS ******
    V1_RETRIES_EXHAUSTED_ERROR = 31
    V1_INVALID_CREDENTIALS_ERROR = 32
    V1_SERVER_DOWN_OR_UNREACHABLE_ERROR = 33
    V1_CHIF_DRIVER_MISSING_ERROR = 34
    REST_ILOREST_CHIF_DLL_MISSING_ERROR = 35
    REST_ILOREST_UNEXPECTED_RESPONSE_ERROR = 36
    REST_ILOREST_ILO_ERROR = 37
    REST_ILOREST_CREATE_BLOB_ERROR = 38
    REST_ILOREST_READ_BLOB_ERROR = 39

    # ****** RDMC ERRORS ******
    SAME_SETTINGS_ERROR = 40
    FIRMWARE_UPDATE_ERROR = 41
    BOOT_ORDER_ENTRY_ERROR = 42
    NIC_MISSING_OR_INVALID_ERROR = 43
    NO_CURRENT_SESSION_ESTABLISHED = 44
    FAILURE_DURING_COMMIT_OPERATION = 45
    USERNAME_PASSWORD_REQUIRED_ERROR = 46
    VNIC_NOT_ENABLED_ERROR = 47
    MULTIPLE_SERVER_CONFIG_FAIL = 51
    MULTIPLE_SERVER_INPUT_FILE_ERROR = 52
    LOAD_SKIP_SETTING_ERROR = 53
    INCOMPATIBLE_ILO_VERSION_ERROR = 54
    INVALID_CLIST_FILE_ERROR = 55
    UNABLE_TO_MOUNT_BB_ERROR = 56
    BIRTHCERT_PARSE_ERROR = 57
    INCOMPATIBLE_SERVER_TYPE = 58
    ILO_LICENSE_ERROR = 59
    RESOURCE_EXISTS_ERROR = 60

    # ****** RMC/RIS ERRORS ******
    RIS_VALUE_CHANGED_ERROR = 61
    RIS_REF_PATH_NOT_FOUND_ERROR = 62
    RIS_ILO_RESPONSE_ERROR = 63
    RIS_ILO_INIT_ERROR = 64
    RIS_SCHEMA_PARSE_ERROR = 65
    RIS_ILO_CHIF_ACCESS_DENIED_ERROR = 66
    RIS_CREATE_AND_PREPARE_CHANNEL_ERROR = 67
    RIS_ILO_CHIF_PACKET_EXCHANGE_ERROR = 71
    RIS_ILO_CHIF_NO_DRIVER_ERROR = 69

    # ****** REST V1 ERRORS ******
    REST_ILOREST_WRITE_BLOB_ERROR = 70
    REST_ILOREST_BLOB_DELETE_ERROR = 68
    REST_ILOREST_BLOB_FINALIZE_ERROR = 72
    REST_ILOREST_BLOB_NOT_FOUND_ERROR = 73
    JSON_DECODE_ERROR = 74
    V1_SECURITY_STATE_ERROR = 75
    REST_ILOREST_BLOB_OVERRIDE_ERROR = 76
    REST_BLOB_RETRIES_EXHAUSETED_ERROR = 77

    # ****** RDMC ERRORS ******
    RESOURCE_ALLOCATION_ISSUES_ERROR = 80
    ENCRYPTION_ERROR = 81
    DRIVE_MISSING_ERROR = 82
    PATH_UNAVAILABLE_ERROR = 83
    ILO_RIS_CORRUPTION_ERROR = 84
    RESOURCE_NOT_READY_ERROR = 85

    # ****** RIS ERRORS ******
    RIS_RIS_BIOS_UNREGISTERED_ERROR = 100

    # ***** Upload/Download ERRORS ******
    FAILED_TO_DOWNLOAD_COMPONENT = 101
    UPDATE_SERVICE_BUSY = 102
    FAILED_TO_UPLOAD_COMPONENT = 103
    TASKQUEUE_ERROR = 104
    DEVICE_DISCOVERY_IN_PROGRESS = 105
    INSTALLSET_ERROR = 106
    INVALID_TARGET_ERROR = 107
    UPLOAD_POWEROFF_ERROR = 108

    # **** ComputeOpsManagement Errors****
    CLOUD_CONNECT_TIMEOUT = 111
    CLOUD_CONNECT_FAILED = 112
    CLOUD_ALREADY_CONNECTED = 113
    PROXY_CONFIG_FAILED = 114

    # **** scep error ****
    SCEP_ENABLED_ERROR = 121

    # *** authentication error ****
    TFA_WRONG_OTP = 131
    TFA_OTP_TIMEDOUT = 132
    TFA_ENABLED_ERROR = 133
    TFA_OTP_EMAILED = 134

    # ****** GENERAL ERRORS ******
    GENERAL_ERROR = 255


class RdmcError(Exception):
    """Baseclass for all rdmc exceptions"""

    errcode = 1

    def __init__(self, message=None):
        Exception.__init__(self, message)


class ConfigurationFileError(RdmcError):
    """Raised when something is wrong in the config file"""

    errcode = 3


class ProxyConfigFailedError(RdmcError):
    """Raised when ComputeOpsManagement connection fails"""

    pass


class CloudConnectTimeoutError(RdmcError):
    """Raised when ComputeOpsManagement connection times out"""

    pass


class CloudConnectFailedError(RdmcError):
    """Raised when ComputeOpsManagement connection fails"""

    pass


class TfaEnablePreRequisiteError(RdmcError):
    """Raised when pre-requisites not met while enabling TFA"""

    pass


class AlreadyCloudConnectedError(RdmcError):
    """Raised when ComputeOpsManagement is already connected"""

    pass


class CommandNotEnabledError(RdmcError):
    """Raised when user tries to invoke a command that isn't enabled"""

    pass


class iLORisCorruptionError(RdmcError):
    """Raised when user tries to invoke a command that isn't enabled"""

    pass


class ResourceNotReadyError(RdmcError):
    """Raised when user tries to invoke a command that isn't enabled"""

    pass


class UsernamePasswordRequiredError(RdmcError):
    """Raised when username and password are required for local chif login"""

    pass


class PathUnavailableError(Exception):
    """Raised when the requested path is unavailable"""

    pass


class InvalidCommandLineError(RdmcError):
    """Raised when user enter incorrect command line arguments"""

    pass


class NoCurrentSessionEstablished(RdmcError):
    """Raised when user enter incorrect command line arguments"""

    pass


class NoChangesFoundOrMadeError(RdmcError):
    """Raised when no changes were found or made on the commit function"""

    pass


class StandardBlobErrorHandler(RdmcError):
    """Raised when error occured for blob operations"""

    pass


class InvalidCommandLineErrorOPTS(RdmcError):
    """Raised when user enter incorrect command line arguments"""

    pass


class InvalidFileInputError(RdmcError):
    """Raised when user enter an invalid file input"""

    pass


class InvalidFileFormattingError(RdmcError):
    """Raised when user enter incorrect load file formatting"""

    pass


class WindowsUserNotAdmin(RdmcError):
    """Raised when user is not running as admin"""

    pass


class NoContentsFoundForOperationError(RdmcError):
    """Raised when no contents were found for the current operation"""

    pass


class InfoMissingEntriesError(RdmcError):
    """Raised when no valid entries for info were found in the current
    instance"""

    pass


class InvalidOrNothingChangedSettingsError(RdmcError):
    """Raised when something is wrong with the settings"""

    pass


class InvalidPasswordLengthError(RdmcError):
    """Raised when password length is invalid"""

    pass


class NoDifferencesFoundError(RdmcError):
    """Raised when no differences are found in the current configuration"""

    pass


class NothingSelectedError(RdmcError):
    """Raised when type selection is reference but none have been provided"""

    pass


class InvalidPropertyError(RdmcError):
    """Raised when one or more properties or attributes are in conflict with current or
    specified configuration"""

    pass


class MultipleServerConfigError(RdmcError):
    """Raised when one or more servers failed to load given configuration"""

    pass


class InvalidMSCfileInputError(RdmcError):
    """Raised when servers input file for load has incorrect parameters"""

    pass


class FirmwareUpdateError(RdmcError):
    """Raised when there is an error while updating firmware"""

    pass


class FailureDuringCommitError(RdmcError):
    """Raised when there is an error during commit"""

    pass


class BootOrderMissingEntriesError(RdmcError):
    """Raised when no entries were found for bios tools"""

    pass


class NicMissingOrConfigurationError(RdmcError):
    """ Raised when no entries are found for given NIC or all NICs are \
     configured or when wrong inputs are presented for NIC entries"""

    pass


class IncompatibleiLOVersionError(RdmcError):
    """Raised when the iLO version is above or below the required \
    version"""

    pass


class IncompatableServerTypeError(RdmcError):
    """Raised when the server type is incompatable with the requested\
    command"""

    pass


class IloLicenseError(RdmcError):
    """Raised when the proper iLO license is not available for a command"""

    pass


class ScepenabledError(RdmcError):
    """Raised when the generation csr or deletion of https cert is issues when scep is enabled"""

    pass


class ResourceExists(RdmcError):
    """Raised when the account to be added already exists"""

    pass


class InvalidCListFileError(RdmcError):
    """Raised when an error occurs while reading the cfilelist \
    within AHS logs"""

    pass


class PartitionMoutingError(RdmcError):
    """Raised when there is error or iLO fails to respond to \
    partition mounting request"""

    pass


class DownloadError(RdmcError):
    """Raised when the component fails to download"""

    pass


class UploadError(RdmcError):
    """Raised when the component fails to download"""

    pass

class UploadPowerOffError(RdmcError):
    """Raised when the server power is On while uploading B- component"""

    pass

class TimeOutError(RdmcError):
    """Raised when the update service times out"""

    pass


class LibHPsrvMissingError(RdmcError):
    """Raised when unable to obtain the libhpsrv handle"""

    pass


class BirthcertParseError(RdmcError):
    """Raised when unable to parse the birthcert"""

    pass


class InvalidKeyError(RdmcError):
    """Raised when an invalid encryption key is used"""

    pass


class UnableToDecodeError(RdmcError):
    """Raised when the file is unable to be decoded using the given key"""

    pass


class UnabletoFindDriveError(RdmcError):
    """Raised when there is an issue finding required label"""

    pass


class TaskQueueError(RdmcError):
    """Raised when there is an issue with the current order of taskqueue"""

    pass


class DeviceDiscoveryInProgress(RdmcError):
    """Raised when device discovery in progress"""

    pass


class InvalidSmartArrayConfigurationError(RdmcError):
    """Raised for an invalid configuration of a smart array controller and/or disk configuration"""

    pass


class FallbackChifUse(RdmcError):
    """Fallback Chif Use"""

    pass


class InstallsetError(RdmcError):
    """Error while deleting Recovery installset"""

    pass


class UI(object):
    """UI class handles all of our printing etc so we have
    consistency across the project"""

    def __init__(self, verbosity=1):
        self.verbosity = verbosity

    def printer(self, data, flush=True, excp=None, verbose_override=False):
        """Print wrapper for stdout vs fileout"""
        if self.verbosity >= 0 or verbose_override:
            if excp:
                sys.stderr.write(str(data))
            else:
                try:
                    sys.stdout.write(str(data))
                    if flush:
                        sys.stdout.flush()
                except IOError:
                    pass

    def command_not_found(self, cmd):
        """Called when command was not found"""
        self.printer(
            ("\nCommand '%s' not found. Use the help command to " "see a list of available commands\n" % cmd),
            excp=True,
        )
        return ReturnCodes.UI_CLI_COMMAND_NOT_FOUND_EXCEPTION

    def command_not_enabled(self, cmd, excp):
        """Called when command has not been enabled"""
        self.printer(("\nCommand '%s' has not been enabled: %s\n" % (cmd, excp)), excp=excp)

    def invalid_commmand_line(self, excp):
        """Called when user entered invalid command line entries"""
        self.printer(("Error: %s\n" % excp), excp=excp)

    def ilo_ris_corruption(self, excp):
        """Called when user entered invalid command line entries"""
        self.printer(("\nError: %s\n" % excp), excp=excp)

    def standard_blob_error(self, excp):
        """Called when user error encountered with blob"""
        self.printer(("\nError: Blob operation failed with error code %s\n" % excp), excp=excp)

    def invalid_file_formatting(self, excp):
        """Called when file formatting is unrecognizable"""
        self.printer(("\nError: %s\n" % excp), excp=excp)

    def user_not_admin(self):
        """Called when file formatting in unrecognizable"""
        self.printer(
            (
                "\nBoth remote and local mode is accessible when %s "
                "is run as administrator. Only remote mode is available for non-"
                "admin user groups.\n" % versioning.__longname__
            ),
            excp=True,
        )

    def no_contents_found_for_operation(self, excp):
        """Called when no contents were found for the current operation"""
        self.printer(("\nError: %s\n" % excp), excp=True)

    def nothing_selected(self):
        """Called when nothing has been select yet"""
        self.printer(
            "\nNo type currently selected. Please use the"
            " 'types' command to\nget a list of types, or input"
            " your type by using the '--selector' flag.\n",
            excp=True,
        )

    def nothing_selected_filter(self):
        """Called when nothing has been select after a filter set"""
        self.printer("\nNothing was found to match your provided filter.\n", excp=True)

    def nothing_selected_set(self):
        """Called when nothing has been select yet"""
        self.printer("\nNothing is selected or selection is read-only.\n", excp=True)

    def no_differences_found(self, excp):
        """Called when no difference is found in the current configuration"""
        self.printer(("Error: %s\n" % excp), excp=True)

    def multiple_server_config_fail(self, excp):
        """Called when one or more servers failed to load given configuration"""
        self.printer(("Error: %s\n" % excp), excp=True)

    def multiple_server_config_input_file(self, excp):
        """Called when servers input file has incorrect information"""
        self.printer(("Error: %s\n" % excp), excp=True)

    def invalid_credentials(self, timeout):
        """Called user has entered invalid credentials
        :param timeout: timeout given for failed login attempt
        :type timeout: int.
        """
        self.printer("Validating...", excp=True)

        for _ in range(0, (int(str(timeout)) + 10)):
            time.sleep(1)
            self.printer(".", excp=True)

        self.printer(
            "\nError: Could not authenticate. Invalid " "credentials, or bad username/password.\n",
            excp=True,
        )

    def bios_unregistered_error(self):
        """Called when ilo/bios unregistered error occurs"""
        self.printer(
            "\nERROR 100: Bios provider is unregistered. Please"
            " refer to the documentation for details on this issue.\n",
            excp=True,
        )

    def error(self, msg, inner_except=None):
        """Used for general error handling
        :param inner_except: raised exception to be logged
        :type inner_except: exception.
        :param msg: warning message
        :type msg: string.
        """
        if inner_except is not None:
            LOGGER.error(msg, exc_info=True)
        else:
            LOGGER.error(msg)

    def warn(self, msg, inner_except=None):
        """Used for general warning handling
        :param inner_except: raised exception to be logged
        :type inner_except: exception.
        :param msg: warning message
        :type msg: string.
        """
        if inner_except is not None:
            LOGGER.warning(msg, exc_info=True)
        else:
            LOGGER.warning(msg)

    def retries_exhausted_attemps(self):
        """Called when url retries have been exhausted"""
        LOGGER.error("Could not reach URL. Retries have been exhausted.")

    def retries_exhausted_vnic_not_enabled(self):
        """Called when there is no VNIC is Enabled"""
        self.printer("\nError: Could not reach URL, VNIC is not enabled. \n", excp=True)

    def print_out_json(self, content):
        """Print out json content to std.out with sorted keys
        :param content: content to be printed out
        :type content: str.
        """
        # stringify
        content = json.dumps(content, indent=2, cls=redfish.ris.JSONEncoder, sort_keys=True)
        self.printer(content, verbose_override=True)
        self.printer("\n")

    def print_out_json_ordered(self, content):
        """Print out sorted json content to std.out
        :param content: content to be printed out
        :type content: str.
        """
        content = json.dumps(content, indent=2, cls=redfish.ris.JSONEncoder)
        self.printer(content, verbose_override=True)
        self.printer("\n")

    def print_out_human_readable(self, content):
        """Print out human readable content to std.out
        :param content: content to be printed out
        :type content: str.
        """
        self.pretty_human_readable(content, enterloop=True)
        self.printer("\n")

    def pretty_human_readable(self, content, indent=0, start=0, enterloop=False):
        """Convert content to human readable and print out to std.out
        :param content: content to be printed out
        :type content: str.
        :param indent: indent string to be used as seperator
        :type indent: str.
        :param start: used to determine the indent level
        :type start: int.
        """
        space = "\n" + "\t" * indent + " " * start
        if isinstance(content, list):
            for item in content:
                if item is None:
                    continue

                self.pretty_human_readable(item, indent, start)

                if content.index(item) != (len(content) - 1):
                    self.printer(space)
        elif isinstance(content, dict):
            for key, value in content.items():
                if space and not enterloop:
                    self.printer(space)

                enterloop = False
                self.printer((str(key) + "="))
                self.pretty_human_readable(value, indent, (start + len(key) + 2))
        else:
            content = content if isinstance(content, six.string_types) else str(content)

            content = '""' if not content else content
            # Changed to support py3, verify if there is a unicode prit issue.

            self.printer(content)


class Encryption(object):
    """Encryption/Decryption object"""

    @staticmethod
    def check_fips_mode_os():
        """Function to check for the OS fips mode
        :param key: string to encrypt with
        :type key: str.
        :returns: returns True if FIPS mode is active, False otherwise
        """
        fips = False
        if os.name == "nt":
            reg = winreg.ConnectRegistry(None, HKEY_LOCAL_MACHINE)
            try:
                reg = winreg.OpenKey(
                    reg,
                    "System\\CurrentControlSet\\Control\\" "Lsa\\FipsAlgorithmPolicy",
                )
                winreg.QueryInfoKey(reg)
                value, _ = winreg.QueryValueEx(reg, "Enabled")
                if value:
                    fips = True
            except:
                fips = False
        else:
            try:
                fipsfile = open("/proc/sys/crypto/fips_enabled")
                result = fipsfile.readline()
                if int(result) > 0:
                    fipsfile = True
                fipsfile.close()
            except:
                fips = False
        return fips

    @staticmethod
    def check_fips_mode_ssl():
        """Function to check for the SSL fips mode
        Uses custom cpython ssl module API, if available. Otheriwse
        probes using ctypes.cdll APIs.
        :returns: returns True if FIPS mode is active, False otherwise
        """
        import ssl

        if hasattr(ssl, "FIPS_mode"):
            return ssl.FIPS_mode()

        from ctypes import cdll

        libcrypto = cdll.LoadLibrary(ssl._ssl.__file__)
        return libcrypto.FIPS_mode()

    def encrypt_file(self, filetxt, key):
        """encrypt a file given a key
        :param filetxt: content to be encrypted
        :type content: str.
        :param key: string to encrypt with
        :type key: str.
        """
        try:
            filetxt = filetxt.encode()
        except (UnicodeDecodeError, AttributeError):
            pass  # must be encoded already
        try:
            key = key.encode()
        except (UnicodeDecodeError, AttributeError):
            pass  # must be encoded already
        if Encryption.check_fips_mode_os():
            raise CommandNotEnabledError("Encrypting of files is not available" " in FIPS mode.")
        if len(key) not in [16, 24, 32]:
            raise InvalidKeyError("")
        else:
            encryptedfile = pyaes.AESModeOfOperationCTR(key).encrypt(filetxt)

        return encryptedfile

    def decrypt_file(self, filetxt, key):
        """decrypt a file given a key
        :param filetxt: content to be decrypted
        :type content: str.
        :param key: string to decrypt with
        :type key: str.
        :returns: returns the decrypted file
        """
        try:
            filetxt = filetxt.encode()
        except (UnicodeDecodeError, AttributeError):
            pass  # must be encoded already
        try:
            key = key.encode()
        except (UnicodeDecodeError, AttributeError):
            pass  # must be encoded already
        if len(key) not in [16, 24, 32]:
            raise InvalidKeyError("")
        else:
            decryptedfile = pyaes.AESModeOfOperationCTR(key).decrypt(filetxt)
            try:
                json.loads(decryptedfile)
            except:
                raise UnableToDecodeError(
                    "Unable to decrypt the file, make " "sure the key is the same as used in encryption."
                )

        return decryptedfile

    @staticmethod
    def decode_credentials(credential):
        """decode an encoded credential
        :param credential: credential to be decoded
        :type credential: str.
        :returns: returns the decoded credential
        """

        lib = risblobstore2.BlobStore2.gethprestchifhandle()
        credbuff = create_string_buffer(credential.encode("utf-8"))
        retbuff = create_string_buffer(128)

        lib.decode_credentials.argtypes = [c_char_p]

        lib.decode_credentials(credbuff, byref(retbuff))

        risblobstore2.BlobStore2.unloadchifhandle(lib)
        # try:
        #    if isinstance(retbuff.value, bytes):
        #        retbuff.value = retbuff.value.decode('utf-8', 'ignore')
        #    if not retbuff.value:
        #        raise UnableToDecodeError("")
        # except:
        #    raise UnableToDecodeError("Unable to decode credential %s." % credential)

        return retbuff.value

    @staticmethod
    def encode_credentials(credential):
        """encode a credential
        :param credential: credential to be encoded
        :type credential: str.
        :returns: returns the encoded credential
        """

        lib = risblobstore2.BlobStore2.gethprestchifhandle()
        if isinstance(credential, bytes):
            credential = credential.decode("utf-8")
        credbuff = create_string_buffer(credential.encode("utf-8"))

        retbuff = create_string_buffer(128)

        lib.encode_credentials.argtypes = [c_char_p]

        lib.encode_credentials(credbuff, byref(retbuff))

        risblobstore2.BlobStore2.unloadchifhandle(lib)
        try:
            if six.PY2:
                enc_val = retbuff.value.encode("utf-8")
            elif six.PY3:
                enc_val = retbuff.value.decode("utf-8")  # .encode('utf-8')
            if not retbuff.value:
                raise UnableToDecodeError("")
        except Exception:
            raise UnableToDecodeError("Unable to decode credential %s." % credential)

        return enc_val


class TabAndHistoryCompletionClass(Completer):
    """Tab and History Class used by interactive mode"""

    def __init__(self, options):
        self.options = options
        self.toolbar_text = None
        self.last_complete = None

    def get_completions(self, document, complete_event):
        """Function to return the options for autocomplete"""
        word = ""
        self.toolbar_text = ""
        lstoption = self.options
        if document.text:
            tokens = document.text.split()
            # We aren't completing options yet
            tokens = [token for token in tokens if not token.startswith("-")]

            self.last_complete = tokens[-1]
            nestedtokens = self.last_complete.split("/")

            if not document.text.endswith(" "):
                tokens.pop()
                word = document.get_word_under_cursor()
            else:
                nestedtokens = []
            if word == "/":
                word = ""

            if len(tokens) >= 1:
                if tokens[0] == "select":
                    # only first type
                    if len(tokens) >= 2:
                        lstoption = []
                    else:
                        lstoption = self.options.get(tokens[0], {})
                elif tokens[0] in ["get", "list", "info", "set"]:
                    # Match properties
                    nested_data = self.options.get("nestedprop", {})
                    nested_info = self.options.get("nestedinfo", {})
                    for token in nestedtokens:
                        try:
                            nested_data = nested_data[token]
                            if tokens[0] == "get" and isinstance(nested_data, dict):
                                for k in list(nested_data.keys()):
                                    if (
                                        k.lower() in HARDCODEDLIST
                                        or "@odata" in k.lower()
                                        or "@redfish.allowablevalues" in k.lower()
                                    ):
                                        del nested_data[k]
                            if nested_info:
                                if "properties" in nested_info:
                                    nested_info = nested_info["properties"]
                                if "AttributeName" not in nested_info[token]:
                                    nested_info = (
                                        nested_info["properties"][token]
                                        if "properties" in nested_info
                                        else nested_info[token]
                                    )
                                else:
                                    nested_info = nested_info[token]
                        except Exception:
                            break
                    nested_data = list(nested_data.keys()) if isinstance(nested_data, dict) else []
                    lstoption = nested_data

                    # Try to get info for help bar
                    help_text = nested_info.get("HelpText", "")
                    enum_tab = []
                    if "Type" in nested_info and nested_info["Type"].lower() == "enumeration":
                        help_text += "\nPossible Values:\n"
                        for value in nested_info["Value"]:
                            enum_tab.append(value["ValueName"])
                            help_text += six.u(str(value["ValueName"])) + " "

                    if not help_text:
                        try:
                            nested_info = nested_info["properties"]
                        except KeyError:
                            pass
                        help_text = nested_info.get("description", "")
                        if "enum" in nested_info:
                            help_text += "\nPossible Values:\n"
                            for value in nested_info["enum"]:
                                enum_tab.append(value)
                                help_text += six.u(str(value)) + " "
                    if isinstance(help_text, str):
                        help_text = help_text.replace(". ", ".\n")
                    self.toolbar_text = help_text
                    if tokens[0] in ["set"]:
                        lstoption = self.options.get("set")
                else:
                    lstoption = {}
            else:
                for token in tokens:
                    # just match commands
                    lstoption = self.options.get(token, {})

        for opt in lstoption:
            if opt == word:
                self.last_complete = opt
            if opt.startswith(word):
                yield Completion(opt + "", start_position=-len(word))

    def bottom_toolbar(self):
        return self.toolbar_text if self.toolbar_text else None

    def updates_tab_completion_lists(self, options):
        """Function to update tab completion lists
        :param options: options list
        :type options: list.
        """
        # Loop through options passed and add them to them
        # to the current tab options list
        for key, value in options.items():
            self.options[key] = value



###
# Copyright 2016-2023 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Firmware Update Command for rdmc """

import time

from redfish.ris.resp_handler import ResponseHandler

try:
    from rdmc_helper import (
        FirmwareUpdateError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        FirmwareUpdateError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )


class FirmwareUpdateCommand:
    """Reboot server that is currently logged in"""

    def __init__(self):
        self.ident = {
            "name": "firmwareupdate",
            "usage": None,
            "description": "Apply a firmware "
            "update to the current logged in server.\n\texample: "
            "firmwareupdate <url/hostname>/images/image.bin\n\t"
            "update based on targets GPU .\n\texample: "
            "firmwareupdate <url/hostname> --targets <id>",
            "summary": "Perform a firmware update on the currently logged in server.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main firmware update worker function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")
        # Third Party FW
        args[0] = args[0].lower()
        if args and (".pup" in args[0] or ".fup" in args[0] or ".hpb" in args[0] or ".HPb" in args[0]):
            raise InvalidCommandLineErrorOPTS("For flashing 3rd party firmwares, please use flashfwpkg command")

        # Supports only .bin/fwpkg/full/vme Firmware files.
        if args and (".bin" in args[0] or ".fwpkg" in args[0] or ".full" in args[0] or ".vme" in args[0]):
            if args[0].startswith('"') and args[0].endswith('"'):
                args[0] = args[0][1:-1]
        else:
            raise InvalidCommandLineErrorOPTS("Please input URL pointing to a .bin or .fwpkg or .full or .vme firmware")

        self.firmwareupdatevalidation(options)

        action = None
        uri = "FirmwareURI"

        select = self.rdmc.app.typepath.defs.hpilofirmwareupdatetype
        results = self.rdmc.app.select(selector=select)

        try:
            results = results[0]
        except:
            pass

        if results:
            update_path = results.resp.request.path
        else:
            raise NoContentsFoundForOperationError("Unable to find %s" % select)

        bodydict = results.resp.dict

        try:
            if options.targets:
                for item in bodydict["Oem"]["Hpe"]["Actions"]:
                    if "AddFromUri" in item:
                        action = item.split("#")[-1]
                    uri = "ImageURI"
                    if action:
                        put_path = bodydict["Oem"]["Hpe"]["Actions"][item]["target"]
                        break
            else:
                for item in bodydict["Actions"]:
                    if self.rdmc.app.typepath.defs.isgen10:
                        if "SimpleUpdate" in item:
                            action = item.split("#")[-1]

                        uri = "ImageURI"
                        # options.tpmenabled = False
                    elif "InstallFromURI" in item:
                        action = "InstallFromURI"

                    if action:
                        put_path = bodydict["Actions"][item]["target"]
                        break
        except:
            put_path = update_path
            action = "Reset"

        if "AddFromUri" in action:
            body = {
                "Action": action,
                uri: args[0],
                "TPMOverrideFlag": options.tpmenabled,
            }
        elif "SimpleUpdate" in action:
            body = {"Action": action, uri: args[0]}
        if options.targets:
            targets = []
            target_list = options.targets.split(",")
            for target in target_list:
                target_url = "/redfish/v1/UpdateService/FirmwareInventory/" + str(target) + "/"
                targets.append(target_url)
            verified = self.verify_targets(options.targets)
            if not verified:
                self.rdmc.ui.error("Provided target was not available, Please provide valid target id\n")
                return ReturnCodes.INVALID_TARGET_ERROR
            body["UpdateTarget"] = options.updatetarget
            body["UpdateRepository"] = options.updaterepository
            body["Targets"] = targets
            self.rdmc.ui.printer('payload: "%s"\n' % body)
        self.rdmc.app.post_handler(put_path, body, silent=True, service=True)

        self.rdmc.ui.printer("\nStarting upgrade process...\n\n")

        self.showupdateprogress(update_path)
        # logoutobj.run("")

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def verify_targets(self, target):
        target_list = target.split(",")
        for target in target_list:
            try:
                target_url = "/redfish/v1/UpdateService/FirmwareInventory/" + target
                dd = self.rdmc.app.get_handler(target_url, service=True, silent=True)
                if dd.status == 404:
                    return False
            except:
                return False
        return True

    def showupdateprogress(self, path):
        """handler function for updating the progress

        :param path: path to update service.
        :tyep path: str
        """
        counter = 0
        written = False
        uploadingpost = False
        spinner = ["|", "/", "-", "\\"]
        position = 0

        while True:
            if counter == 100:
                raise FirmwareUpdateError(
                    "Error occurred while updating the firmware. Check if TPM is enabled. "
                    "And if TPM enabled provide --tpmenabled flag"
                )
            else:
                counter += 1

            results = self.rdmc.app.get_handler(path, silent=True)
            results = results.dict
            try:
                results = results["Oem"]["Hpe"]
            except:
                pass

            if not results:
                raise FirmwareUpdateError("Unable to contact Update Service. " "Please re-login and try again.")

            if results["State"].lower().startswith("idle"):
                time.sleep(2)
            elif results["State"].lower().startswith("uploading"):
                counter = 0

                if not uploadingpost:
                    uploadingpost = True
                else:
                    if not written:
                        written = True
                        self.rdmc.ui.printer("\n iLO is uploading the necessary files. Please wait...\n")

                time.sleep(0.5)
            elif results["State"].lower().startswith(("progressing", "updating", "verifying", "writing")):
                counter = 0

                for _ in range(2):
                    if position < 4:
                        self.rdmc.ui.printer("Updating: " + spinner[position] + "\r")
                        position += 1
                        time.sleep(0.1)
                    else:
                        position = 0
                        self.rdmc.ui.printer("Updating: " + spinner[position] + "\r")
                        position += 1
                        time.sleep(0.1)
            elif results["State"].lower().startswith("complete"):
                self.rdmc.ui.printer(
                    "\n\nFirmware update has completed and iLO"
                    " may reset. \nIf iLO resets the"
                    " session will be terminated.\nPlease wait"
                    " for iLO to initialize completely before"
                    " logging in again.\nA reboot may be required"
                    " for firmware changes to take effect.\n"
                )
                break
            elif results["State"].lower().startswith("error"):
                error = self.rdmc.app.get_handler(path, silent=True)
                self.printerrmsg(error)

    def printerrmsg(self, error):
        """raises and prints the detailed error message if possible"""
        output = "Error occurred while updating the firmware."

        try:
            error = error.dict["Oem"]["Hpe"]["Result"]["MessageId"].split(".")
            # TODO: Update to new ResponseHandler Method 'return_reg'
            errmessages = ResponseHandler(
                self.rdmc.app.validation_manager,
                self.rdmc.app.typepath.defs.messageregistrytype,
            ).get_error_messages()
            for messagetype in list(errmessages.keys()):
                if error[0] == messagetype:
                    if errmessages[messagetype][error[-1]]["NumberOfArgs"] == 0:
                        output = "Firmware update error. %s" % errmessages[messagetype][error[-1]]["Message"]
                    else:
                        output = "Firmware update error. %s" % errmessages[messagetype][error[-1]]["Description"]
                    break
        except:
            pass

        raise FirmwareUpdateError(output)

    def firmwareupdatevalidation(self, options):
        """Firmware update method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--tpmenabled",
            dest="tpmenabled",
            action="store_true",
            help="Use this flag if the server you are currently logged into" " has a TPM chip installed.",
            default=False,
        )
        customparser.add_argument(
            "--updatetarget",
            dest="updatetarget",
            action="store_true",
            help="UpdateTarget set as true/false" "if true firmware updates",
            default=True,
        )
        customparser.add_argument(
            "--updaterepository",
            dest="updaterepository",
            action="store_false",
            help="UpdateRepository set as true or false" " if true it will add repository.",
            default=True,
        )
        customparser.add_argument(
            "--targets",
            help="If targets value specify a comma separated\t" "firmwareinventory id only",
            metavar="targets_indices",
        )



###
# Copyright 2017 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Directory Command for rdmc """

import re
import sys
from argparse import Action, RawDescriptionHelpFormatter

from redfish.ris.rmc_helper import IdTokenError, IloResponseError

try:
    from rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ResourceExists,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ResourceExists,
        ReturnCodes,
    )

__subparsers__ = ["ldap", "kerberos", "test"]

PRIVKEY = {
    1: ("Login", "AssignedPrivileges"),
    2: ("RemoteConsolePriv", "OemPrivileges"),
    3: ("ConfigureUsers", "AssignedPrivileges"),
    4: ("ConfigureManager", "AssignedPrivileges"),
    5: ("VirtualMediaPriv", "OemPrivileges"),
    6: ("VirtualPowerAndResetPriv", "OemPrivileges"),
    7: ("HostNICConfigPriv", "OemPrivileges"),
    8: ("HostBIOSConfigPriv", "OemPrivileges"),
    9: ("HostStorageConfigPriv", "OemPrivileges"),
    10: ("SystemRecoveryConfigPriv", "OemPrivileges"),
    11: ("ConfigureSelf", "AssignedPrivileges"),
    12: ("ConfigureComponents", "AssignedPrivileges"),
}


class _DirectoryParse(Action):
    def __init__(self, option_strings, dest, nargs, **kwargs):
        super(_DirectoryParse, self).__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_strings):
        """Helper for parsing options"""
        if option_strings.endswith("disable"):
            setattr(namespace, self.dest, False)
        elif option_strings.endswith("enable"):
            setattr(namespace, self.dest, True)
        elif option_strings.endswith("enablelocalauth"):
            setattr(namespace, self.dest, False)
        elif option_strings.endswith("disablelocalauth"):
            setattr(namespace, self.dest, True)
        elif option_strings == "--removerolemap":
            setattr(namespace, self.dest, {"remove": []})
            for role in next(iter(values)).split("#"):
                role = role.replace('"', "")
                if role:
                    namespace.roles["remove"].append(role)
        elif option_strings == "--addrolemap":
            setattr(namespace, self.dest, {"add": []})
            for role in next(iter(values)).split("#"):
                role = role.replace('"', "")
                if role and re.match(".*:.*", role):
                    privs = role.split(":")[0].split(";")
                    if len(privs) > 1:
                        for priv in privs:
                            try:
                                if priv and int(priv) > 12:
                                    try:
                                        parser.error("Invalid privilege number added %s." % priv)
                                    except SystemExit:
                                        raise InvalidCommandLineErrorOPTS("")
                            except ValueError:
                                try:
                                    parser.error("Privileges must be added as numbers.")
                                except SystemExit:
                                    raise InvalidCommandLineErrorOPTS("")
                    namespace.roles["add"].append(role)
                else:
                    try:
                        parser.error("Supply roles to add in form <local role>:<remote group>")
                    except SystemExit:
                        raise InvalidCommandLineErrorOPTS("")
        elif option_strings == "--addsearch":
            setattr(namespace, self.dest, {"add": []})
            for search in next(iter(values)).split(";"):
                if search:
                    namespace.search["add"].append(search)

        elif option_strings == "--removesearch":
            setattr(namespace, self.dest, {"remove": []})
            for search in next(iter(values)).split(";"):
                if search:
                    namespace.search["remove"].append(search)


class DirectoryCommand:
    """Update directory settings on the server"""

    def __init__(self):
        self.ident = {
            "name": "directory",
            "usage": None,
            "description": "\tAdd credentials, service address, two search strings, and enable"
            "\n\tLDAP directory service, remote role groups (mapping), local custom role\n\t"
            "IDs with privileges.\n\n\tTo view help on specific sub-commands"
            " run: directory <sub-command> -h\n\n\tExample: directory ldap -h\n",
            "summary": "Update directory settings, add/delete directory roles, and test directory "
            "settings on the currently logged in server.",
            "aliases": ["ad", "activedirectory"],
            "auxcommands": ["IloAccountsCommand"],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main directory Function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            ident_subparser = False
            for cmnd in __subparsers__:
                if cmnd in line:
                    (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
                    ident_subparser = True
                    break
            if not ident_subparser:
                (options, args) = self.rdmc.rdmc_parse_arglist(self, line, default=True)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.directoryvalidation(options)

        if self.rdmc.app.getiloversion() < 5.140:
            raise IncompatibleiLOVersionError("Directory settings are only available on " "iLO 5 1.40 or greater.")
        results = None
        if options.command.lower() == "ldap" or (
            (True if options.ldap_kerberos == "ldap" else False) if hasattr(options, "ldap_kerberos") else False
        ):
            try:
                results = self.rdmc.app.select(selector="AccountService.", path_refresh=True)[0].dict
                path = results[self.rdmc.app.typepath.defs.hrefstring]
                oem = results["Oem"][self.rdmc.app.typepath.defs.oemhp]
                local_auth = results["LocalAccountAuth"]
                results = results["LDAP"]
                if "RemoteRoleMapping" not in results:
                    rolemap = {"RemoteRoleMapping": {}}
                    results.update(rolemap)
                name = "LDAP"
            except (KeyError, IndexError):
                raise NoContentsFoundForOperationError("Unable to gather LDAP settings.")

        elif options.command.lower() == "kerberos" or (
            (True if options.ldap_kerberos == "kerberos" else False) if hasattr(options, "ldap_kerberos") else False
        ):
            try:
                results = self.rdmc.app.select(selector="AccountService.", path_refresh=True)[0].dict
                path = results[self.rdmc.app.typepath.defs.hrefstring]
                oem = results["Oem"][self.rdmc.app.typepath.defs.oemhp]
                local_auth = results["LocalAccountAuth"]
                results = results["ActiveDirectory"]
                if "RemoteRoleMapping" not in results:
                    rolemap = {"RemoteRoleMapping": {}}
                    results.update(rolemap)
                name = "ActiveDirectory"
            except (KeyError, IndexError):
                raise NoContentsFoundForOperationError("Unable to gather Kerberos settings.")

        if results:
            keytab = None
            payload = {}
            if hasattr(options, "keytab"):
                keytab = options.keytab
            try:
                directory_settings = self.directory_helper(results, options)
                if "ldap" in line and options.roles and "add" in options.roles:
                    role_val = options.roles["add"]
                    for r in role_val:
                        if ";" in r.split(":")[0]:
                            name = "ActiveDirectory"
                            break
            except IndexError:
                directory_settings = self.directory_helper(results, options)

            if directory_settings:
                payload[name] = directory_settings

            if hasattr(options, "authmode"):
                if options.authmode:
                    payload.update(
                        {"Oem": {"Hpe": {"DirectorySettings": {"LdapAuthenticationMode": options.authmode}}}}
                    )

            if not payload and not keytab:
                if getattr(options, "json", False):
                    # self.rdmc.ui.print_out_json({name: results, 'LocalAccountAuth': local_auth,
                    #                             "Oem": {"Hpe": oem}})
                    text_content = self.print_s(results, oem, local_auth, name)
                    self.rdmc.ui.print_out_json(text_content)
                else:
                    self.print_settings(results, oem, local_auth, name)

            if payload:
                priv_patches = {}
                try:
                    if hasattr(options, "localauth"):
                        if options.localauth:
                            payload["LocalAccountAuth"] = "Enabled" if options.localauth else "Disabled"
                    elif local_auth:
                        payload["LocalAccountAuth"] = "Enabled" if local_auth else "Disabled"
                except (NameError, AttributeError):
                    payload["LocalAccountAuth"] = "Disabled"
                try:
                    maps = {}
                    if payload.get("LDAP"):
                        maps = payload["LDAP"].get("RemoteRoleMapping", {})
                    elif payload.get("ActiveDirectory"):
                        maps = payload["ActiveDirectory"].get("RemoteRoleMapping", {})
                        # Check if we need to modify roles after creating
                        for mapping in maps:
                            privs = mapping["LocalRole"].split(";")
                            if len(privs) > 1:
                                privs = [int(priv) for priv in privs if priv]

                                if 10 in privs:
                                    user_privs = self.auxcommands["iloaccounts"].getsesprivs()
                                    if "SystemRecoveryConfigPriv" not in list(user_privs.keys()):
                                        raise IdTokenError(
                                            "The currently logged in account "
                                            "must have the System Recovery Config privilege to "
                                            "add the System Recovery Config privilege to a local "
                                            "role group."
                                        )

                                priv_patches[mapping["RemoteGroup"]] = privs
                                mapping["LocalRole"] = "ReadOnly"
                except Exception as excp:
                    self.rdmc.ui.error(excp)
                self.rdmc.ui.printer("Changing settings...\n")
                try:
                    self.rdmc.app.patch_handler(path, payload)
                except IloResponseError as excp:
                    if not results["ServiceEnabled"]:
                        self.rdmc.ui.error(
                            "You must enable this directory service before or "
                            "during assignment of username and password. Try adding the flag "
                            "--enable.\n",
                            excp,
                        )
                    else:
                        raise IloResponseError
                if priv_patches:
                    self.update_mapping_privs(priv_patches)
            if keytab:
                path = oem["Actions"][next(iter(oem["Actions"]))]["target"]
                self.rdmc.ui.printer("Adding keytab...\n")
                self.rdmc.app.post_handler(path, {"ImportUri": keytab})
        elif options.command.lower() == "test":
            self.test_directory(options, json=getattr(options, "json", False))
        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def update_mapping_privs(self, roles_to_update):
        """Helper function to update created role mappings to match user privileges.

        :param roles_to_update: Dictionary of privileges to update.
        :type roles_to_update: dict
        """
        self.rdmc.ui.printer("Updating privileges of created role maps...\n")
        try:
            results = self.rdmc.app.select(selector="AccountService.", path_refresh=True)[0].dict
            roles = self.rdmc.app.getcollectionmembers(self.rdmc.app.getidbytype("RoleCollection.")[0])
        except (KeyError, IndexError):
            raise NoContentsFoundForOperationError(
                "Unable to gather Role settings. Roles may not " "be updated to match privileges requested."
            )
        for rolemap in results["LDAP"]["RemoteRoleMapping"]:
            for role in roles:
                if role["RoleId"] == rolemap["LocalRole"]:
                    role["RemoteGroup"] = rolemap["RemoteGroup"]
                    break
        for role in roles:
            privs = {"AssignedPrivileges": [], "OemPrivileges": []}

            for update_role in list(roles_to_update.keys()):
                if role.get("RemoteGroup", None) == update_role:
                    for priv in roles_to_update[update_role]:
                        privs[PRIVKEY[priv][1]].append(PRIVKEY[priv][0])
                    try:
                        self.rdmc.app.patch_handler(role["@odata.id"], privs)
                        self.rdmc.ui.printer("Updated privileges for %s\n" % update_role)
                    except IloResponseError as excp:
                        self.rdmc.ui.error("Unable to update privileges for %s\n" % update_role, excp)
                    break

    def directory_helper(self, settings, options):
        """Helper function to set the payload based on options and arguments

        :param settings: dictionary to change
        :type settings: dict.
        :param options: list of options
        :type options: list.
        """

        payload = {}
        serviceaddress = None

        if hasattr(options, "serviceaddress"):
            if isinstance(options.serviceaddress, str):
                serviceaddress = options.serviceaddress
                if serviceaddress == '""' or serviceaddress == "''":
                    serviceaddress = ""
        if hasattr(options, "port"):
            if isinstance(options.port, str):
                if serviceaddress is None:
                    serviceaddress = settings["ServiceAddresses"][0]
                serviceaddress = serviceaddress + ":" + options.port
        if hasattr(options, "realm"):
            if isinstance(options.realm, str):
                if serviceaddress is None:
                    serviceaddress = settings["ServiceAddresses"][0]
                if options.realm == '""' or options.realm == "''":
                    options.realm = ""
                serviceaddress = serviceaddress + "@" + options.realm
        if serviceaddress is not None:
            payload["ServiceAddresses"] = [serviceaddress]

        if hasattr(options, "enable"):
            if options.enable is not None:
                payload["ServiceEnabled"] = options.enable

        if hasattr(options, "ldap_username") and hasattr(options, "ldap_password"):
            if options.ldap_username and options.ldap_password:
                payload.update(
                    {
                        "Authentication": {
                            "Username": options.ldap_username,
                            "Password": options.ldap_password,
                        }
                    }
                )

        if hasattr(options, "roles"):
            if options.roles:
                payload["RemoteRoleMapping"] = self.role_helper(options.roles, settings["RemoteRoleMapping"])

        if hasattr(options, "search"):
            if options.search:
                payload.update(
                    {
                        "LDAPService": {
                            "SearchSettings": self.search_helper(
                                options.search,
                                settings["LDAPService"]["SearchSettings"],
                            )
                        }
                    }
                )
        return payload

    def test_directory(self, options, json=False):
        """Function to perform directory testing

        :param options: namespace of custom parser attributes which contain the original command
                        arguments for 'start/stop/viewresults'
        :type options: namespace
        :param json: Bool to print in json format or not.
        :type json: bool.
        """
        results = self.rdmc.app.select(selector="HpeDirectoryTest.", path_refresh=True)[0].dict
        if options.start_stop_view.lower() == "start":
            path = None
            for item in results["Actions"]:
                if "StartTest" in item:
                    path = results["Actions"][item]["target"]
                    break
            if not path:
                raise NoContentsFoundForOperationError("Unable to start directory test.")
            self.rdmc.ui.printer(
                "Starting the directory test. Monitor results with " 'command: "directory test viewresults".\n'
            )
            self.rdmc.app.post_handler(path, {})
        elif options.start_stop_view.lower() == "stop":
            path = None
            for item in results["Actions"]:
                if "StopTest" in item:
                    path = results["Actions"][item]["target"]
                    break
            if not path:
                raise NoContentsFoundForOperationError("Unable to stop directory test.")
            self.rdmc.ui.printer("Stopping the directory test.\n")
            self.rdmc.app.post_handler(path, {})
        elif options.start_stop_view.lower() == "viewresults":
            if getattr(options, "json", False):
                self.rdmc.ui.print_out_json(results["TestResults"])
            else:
                for test in results["TestResults"]:
                    self.rdmc.ui.printer("Test: %s\n" % test["TestName"])
                    self.rdmc.ui.printer("------------------------\n")
                    self.rdmc.ui.printer("Status: %s\n" % test["Status"])
                    self.rdmc.ui.printer("Notes: %s\n\n" % test["Notes"])

    def print_s(self, settings, oem_settings, local_auth_setting, name):
        setting = "Kerberos" if name == "ActiveDirectory" else name
        enable = str(settings["ServiceEnabled"])
        serviceaddress = settings["ServiceAddresses"][0]
        address = serviceaddress if serviceaddress else "Not Set"
        username = settings["Authentication"]["Username"]
        dis_name = username if username else "Not Set"
        auth = local_auth_setting
        content = {
            "Settings": setting,
            "Enabled": enable,
            "Service Address": address,
            "Distinguished Name": dis_name,
            "Local Account Authorization": auth,
        }
        if name.lower() == "activedirectory":
            address_settings = oem_settings["KerberosSettings"]
            port = address_settings["KDCServerPort"]
            realm = address_settings["KerberosRealm"] if address_settings["KerberosRealm"] else "Not Set"
            content.update({"Port": port, "Realm": realm})

        else:
            address_settings = oem_settings["DirectorySettings"]
            port = address_settings["LdapServerPort"]
            authmode = address_settings["LdapAuthenticationMode"]
            content.update({"Port": port, "Authentication Mode": authmode})
            addsearch = []
            try:
                count = 1
                for search in settings["LDAPService"]["SearchSettings"]["BaseDistinguishedNames"]:
                    # search = (count, search)
                    search = "Search %s: %s" % (count, search)
                    addsearch.append(search)
                    count += 1
                content.update({"Search Settings": addsearch})
            except KeyError:
                search = "No Search Settings"
                content.update({"Search Settings": {search}})
        addval = []
        if "RemoteRoleMapping" in settings:
            if len(settings["RemoteRoleMapping"]) > 0:
                for role in settings["RemoteRoleMapping"]:
                    roleadd = {
                        "Local Role": role["LocalRole"],
                        "Remote Group": role["RemoteGroup"],
                    }
                    addval.append(roleadd)
                content.update({"Remote Role Mapping(s)": addval})
            else:
                addval = {"No Remote Role Mappings"}
                content.update({"Remote Role Mapping(s)": addval})
        else:
            content.update({"No Remote Role Mappings"})
        return content

    def print_settings(self, settings, oem_settings, local_auth_setting, name):
        """Pretty print settings of LDAP or Kerberos

        :param settings: settings to print
        :type settings: dict.
        :param oem_settings: oem_settings to print
        :type oem_settings: dict.
        :param local_auth_settings: local authorization setting
        :type local_auth_settings: str.
        :param name: type of setting (activedirectory or ldap)
        :type name: str.
        """

        self.rdmc.ui.printer("%s settings:\n" % ("Kerberos" if name == "ActiveDirectory" else name))
        self.rdmc.ui.printer("--------------------------------\n")
        self.rdmc.ui.printer("Enabled: %s\n" % str(settings["ServiceEnabled"]))

        serviceaddress = settings["ServiceAddresses"][0]

        self.rdmc.ui.printer("Service Address: %s\n" % (serviceaddress if serviceaddress else "Not Set"))
        username = settings["Authentication"]["Username"]
        self.rdmc.ui.printer("Distinguished Name: %s\n" % (username if username else "Not Set"))

        self.rdmc.ui.printer("Local Account Authorization: %s\n" % local_auth_setting)

        if name.lower() == "activedirectory":
            address_settings = oem_settings["KerberosSettings"]
            self.rdmc.ui.printer("Port: %s\n" % address_settings["KDCServerPort"])

            self.rdmc.ui.printer(
                "Realm: %s\n" % (address_settings["KerberosRealm"] if address_settings["KerberosRealm"] else "Not Set")
            )
        else:
            address_settings = oem_settings["DirectorySettings"]
            self.rdmc.ui.printer("Port: %s\n" % address_settings["LdapServerPort"])
            self.rdmc.ui.printer("Authentication Mode: %s\n" % address_settings["LdapAuthenticationMode"])

            try:
                self.rdmc.ui.printer("Search Settings:\n")
                count = 1
                for search in settings["LDAPService"]["SearchSettings"]["BaseDistinguishedNames"]:
                    self.rdmc.ui.printer("\tSearch %s: %s\n" % (count, search))
                    count += 1
            except KeyError:
                self.rdmc.ui.printer("\tNo Search Settings\n")

        if "RemoteRoleMapping" in settings:
            self.rdmc.ui.printer("Remote Role Mapping(s):\n")
            if len(settings["RemoteRoleMapping"]) > 0:
                for role in settings["RemoteRoleMapping"]:
                    self.rdmc.ui.printer("\tLocal Role: %s\n" % role["LocalRole"])
                    self.rdmc.ui.printer("\tRemote Group: %s\n" % role["RemoteGroup"])
            else:
                self.rdmc.ui.printer("\tNo Remote Role Mappings.\n")
        else:
            self.rdmc.ui.printer("No Remote Role Mappings.\n")

    def role_helper(self, new_roles, curr_roles):
        """Helper to prepare adding and removing roles for patching

        :param new_roles: dictionary of new roles to add or remove
        :type new_roles: dict.
        :param curr_roles: list of current roles on the system
        :type curr_roles: list.
        """
        final_roles = curr_roles
        if "add" in new_roles:
            for role in new_roles["add"]:
                role = role.split(":", 1)
                if not self.duplicate_group(role[1], curr_roles):
                    if len(curr_roles) > 0:
                        final_roles.append({"LocalRole": role[0], "RemoteGroup": role[1]})
                    else:
                        final_roles = [{"LocalRole": role[0], "RemoteGroup": role[1]}]
                        curr_roles = final_roles

                else:
                    raise ResourceExists('Group DN "%s" already exists.' % role[1].split(":")[0])
        if "remove" in new_roles:
            removed = False
            for role in new_roles["remove"]:
                removed = False
                for item in reversed(final_roles):
                    if item["LocalRole"] == role:
                        if len(curr_roles) > 1:
                            del final_roles[final_roles.index(item)]
                            removed = True
                            break
                        elif len(curr_roles) == 1:
                            del final_roles[final_roles.index(item)]
                            final_roles = [{}]
                            removed = True
                            break
                if not removed:
                    raise InvalidCommandLineError("Unable to find local role %s to delete" % role)
        return final_roles

    def duplicate_group(self, group_dn, curr_roles):
        """Checks if new role is a duplicate

        :param group_dn: group domain name from user
        :type group_dn: str.
        :param curr_roles: list of current roles
        :type curr_roles: list.
        """
        group_dn = group_dn.split(":")[0]
        for item in curr_roles:
            comp_dn = item["RemoteGroup"].split(":")[0]
            if comp_dn == group_dn:
                return True
        return False

    def search_helper(self, new_searches, curr_searches):
        """Helper to prepare search strings for patching

        :param new_serches: dictionary of new searches to add
        :type new_searches: dict.
        :param curr_searches: list of current searches
        :type curr_searches: dict.
        """
        final_searches = curr_searches

        if "add" in new_searches:
            if "BaseDistinguishedNames" in final_searches:
                for search in new_searches["add"]:
                    for key, value in final_searches.items():
                        for v in value:
                            if search.lower() == v.lower():
                                raise ResourceExists('Search Setting "%s" already exists.\n' % search)

                    final_searches["BaseDistinguishedNames"].append(search)
            else:
                final_searches["BaseDistinguishedNames"] = new_searches["add"]
        elif "remove" in new_searches:
            to_remove = []

            if "BaseDistinguishedNames" not in curr_searches:
                raise NoContentsFoundForOperationError("No search strings to remove")

            for search in new_searches["remove"]:
                if search in curr_searches["BaseDistinguishedNames"]:
                    to_remove.append(search)
                else:
                    raise InvalidCommandLineError("Unable to find search %s to delete" % search)
            for item in to_remove:
                final_searches["BaseDistinguishedNames"].remove(item)

            if not final_searches["BaseDistinguishedNames"]:
                sys.stdout.write("Attempting to delete all searches.\n")
                final_searches["BaseDistinguishedNames"].append("")

        return final_searches

    def directoryvalidation(self, options):
        """directory validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def options_argument_group(self, parser):
        """Additional argument

        :param parser: The parser to add the removeprivs option group to
        :type parser: ArgumentParser/OptionParser
        """

        parser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        subcommand_parser = customparser.add_subparsers(dest="command")
        default_parser = subcommand_parser.add_parser("default")

        default_parser.add_argument(
            "ldap_kerberos",
            help="Specify LDAP or Kerberos configuration settings",
            metavar="LDAP_KERBEROS",
            nargs="?",
            type=str,
            default=None,
        )
        self.cmdbase.add_login_arguments_group(default_parser)

        privilege_help = (
            "\n\nPRIVILEGES:\n\t1: Login\n\t2: Remote Console\n\t"
            "3: User Config\n\t4: iLO (Manager) Config\n\t5: Virtual Media\n\t"
            "6: Virtual Power and Reset\n\t7: Host NIC Config\n\t8: Host Bios Config\n\t9: "
            "Host Storage Config\n\t10: System Recovery Config\n\t11: Self Password Change\n\t"
            "12: Configure Components\n\n\tLOCAL ROLES:\n\tReadOnly\n\tOperator\n\tAdministrator"
            "\n\n\tNOTE: The Self Password Change privilege is automatically added to roles with "
            "the Login privilege."
        )
        ldap_help = "Show, add or modify properties pertaining to iLO LDAP Configuration."
        ldap_parser = subcommand_parser.add_parser(
            __subparsers__[0],
            help=ldap_help,
            description=ldap_help + "\n\n\tSimply show LDAP configuration:\n\t\tdirectory ldap\n\n"
            "To modify the LDAP username, password, service address, search strings or "
            "enable/disable LDAP.\n\t\tdirectory ldap <username> <password> "
            '--serviceaddress x.x.y.z --addsearch "string1;string2" --enable.\n\n\tTo add role '
            'mapping.\n\t\tdirectory ldap <username> <password> --addrolemap "LocalRole1:'
            'RemoteGroup3#LocalRole2:RemoteGroup4:SID".\n\n\tTo remove role mapping.\n\t\t'
            "directory ldap <username> <password> --removerolemap LocalRole1#LocalRole2." + privilege_help,
            formatter_class=RawDescriptionHelpFormatter,
        )
        ldap_parser.add_argument(
            "ldap_username",
            help="The LDAP username used in verifying AD (optional outside of '--enable' and" "'--disable')",
            metavar="USERNAME",
            nargs="?",
            type=str,
            default=None,
        )
        ldap_parser.add_argument(
            "ldap_password",
            help="The LDAP password used in verifying AD (optional outside of '--enable' and" "'--disable')",
            metavar="PASSWORD",
            nargs="?",
            type=str,
            default=None,
        )
        ldap_parser.add_argument(
            "--enable",
            "--disable",
            dest="enable",
            type=str,
            nargs="*",
            action=_DirectoryParse,
            help="Optionally add this flag to enable LDAP services.",
            default=None,
        )
        ldap_parser.add_argument(
            "--addsearch",
            "--removesearch",
            dest="search",
            nargs="*",
            action=_DirectoryParse,
            help="Optionally add this flag to add or remove search strings for " "generic LDAP services.",
            type=str,
            default={},
        )
        ldap_parser.add_argument(
            "--serviceaddress",
            dest="serviceaddress",
            help="Optionally include this flag to set the service address of the LDAP Services.",
            default=None,
        )
        ldap_parser.add_argument(
            "--port",
            dest="port",
            help="Optionally include this flag to set the port of the LDAP services.",
            default=None,
        )
        ldap_parser.add_argument(
            "--addrolemap",
            "--removerolemap",
            dest="roles",
            nargs="*",
            action=_DirectoryParse,
            help="Optionally add this flag to add or remove Role Mapping(s) for the LDAP."
            " Remove EX: --removerolemap LocalRole1#LocalRole2 "
            'Add EX: --addrolemap "LocalRole1:RemoteGroup3#LocalRole2:RemoteGroup4\n\n"'
            'SID EX: --addrolemap "LocalRole1:RemoteGroup2:SID#LocalRole2:RemoteGroup5:SID"'
            "\n\nNOTE 1: Create a custom local role group (and subsequently assign to a role map)"
            "by adding the numbers associated with privilege(s) desired separated by a semicolon"
            "(;)\n\nNOTE 2: SID is optional",
            type=str,
            default={},
        )
        ldap_parser.add_argument(
            "--enablelocalauth",
            "--disablelocalauth",
            dest="localauth",
            nargs="*",
            type=str,
            action=_DirectoryParse,
            help="Optionally include this flag if you wish to enable or disable authentication " "for local accounts.",
            default=None,
        )
        ldap_parser.add_argument(
            "--authentication",
            dest="authmode",
            choices=["DefaultSchema", "ExtendedSchema"],
            help="Optionally include this flag if you would like to choose a LDAP authentication "
            "mode Valid choices are: DefaultSchema (Directory Default Schema or Schema-free) or "
            "ExtendedSchema (HPE Extended Schema).",
            default=None,
        )
        self.cmdbase.add_login_arguments_group(ldap_parser)

        self.options_argument_group(ldap_parser)

        kerberos_help = "Show, add or modify properties pertaining to AD Kerberos Configuration."
        kerberos_parser = subcommand_parser.add_parser(
            __subparsers__[1],
            help=kerberos_help,
            description=kerberos_help + "\n\nExamples:\n\nShow Kerberos specific AD/LDAP configuration "
            "settings.\n\tdirectory kerberos\n\nShow current AD Kerberos configuration."
            "\n\tdirectory kerberos\n\nAlter kerberos service address, AD realm and Port.\n\t"
            "directory kerberos --serviceaddress x.x.y.z --port 8888 --realm adrealm1",
            formatter_class=RawDescriptionHelpFormatter,
        )
        kerberos_parser.add_argument(
            "--serviceaddress",
            dest="serviceaddress",
            help="Optionally include this flag to set the Kerberos serviceaddress.",
            default=None,
        )
        kerberos_parser.add_argument(
            "--port",
            dest="port",
            help="Optionally include this flag to set the Kerberos port.",
            default=None,
        )
        kerberos_parser.add_argument(
            "--realm",
            dest="realm",
            help="Optionally include this flag to set the Kerberos realm.",
            default=None,
        )
        kerberos_parser.add_argument(
            "--keytab",
            dest="keytab",
            help="Optionally include this flag to import a Kerberos Keytab by it's URI location.",
            default="",
        )
        kerberos_parser.add_argument(
            "--enable",
            "--disable",
            dest="enable",
            type=str,
            nargs="*",
            action=_DirectoryParse,
            help="Optionally add this flag to enable or disable Kerberos services.",
            default=None,
        )
        self.cmdbase.add_login_arguments_group(kerberos_parser)

        self.options_argument_group(kerberos_parser)

        directory_test_help = (
            "Start, stop or view results of an AD/LDAP test which include: ICMP, "
            "Domain Resolution, Connectivity, Authentication, Bindings, LOM Object and User "
            "Context tests."
        )
        directory_test_parser = subcommand_parser.add_parser(
            __subparsers__[2],
            help=directory_test_help,
            description=directory_test_help + "\n\nExamples:\n\nStart a directory test:\n\tdirectory test "
            "start\n\nStop a directory test:\n\tdirectory test stop\n\nView results of the last "
            "directory test:\n\tdirectory test viewresults",
            formatter_class=RawDescriptionHelpFormatter,
        )
        directory_test_parser.add_argument(
            "start_stop_view",
            help="Start, stop, or view results on an AD/LDAP test.",
            metavar="START, STOP, VIEWRESULTS",
            default="viewresults",
        )
        self.cmdbase.add_login_arguments_group(directory_test_parser)
